import time
import asyncio
import random
from typing import Tuple, List, Optional, Literal
from .Data import WINDOWS_KEY_SET, WindowBounds, TouchPoint, KeyModifiers, KeyEvents


class Actions:
    def __init__(self, page_instance):
        self.page_instance = page_instance


# region [ |>*<|=== Domains ===|>*<| ] Input [ |>*<|=== Domains ===|>*<| ]
    #   https://chromedevtools.github.io/devtools-protocol/tot/Input
    #   https://godoc.org/github.com/unixpickle/muniverse/chrome

    async def DispatchKeyEvent(
        self, type_: str,
        modifiers:             int = 0,
        timestamp:             Optional[int] = None,
        text:                  str = "",
        unmodifiedText:        str = "",
        keyIdentifier:         str = "",
        code:                  str = "",
        key:                   str = "",
        windowsVirtualKeyCode: int = 0,
        nativeVirtualKeyCode:  int = 0,
        autoRepeat:           bool = False,
        isKeypad:             bool = True,
        isSystemKey:          bool = False,
        location:              int = 0,
        commands:             Optional[list] = None
    ) -> None:
        """
        Отправляет событие кнопки на страницу.
        https://chromedevtools.github.io/devtools-protocol/tot/Input#method-dispatchKeyEvent
        :param type_:                   Тип события. Допустимые значения:
                                            keyDown, keyUp, rawKeyDown, char
        :param modifiers:               (optional) Битовое поле, представляющее нажатые клавиши-модификаторы.
                                            Alt = 1, Ctrl = 2, Meta/Command = 4, Shift = 8 (по умолчанию: 0).
        :param timestamp:               (optional) UNIX-time (число секунд с 1 января 1970)
        :param text:                    (optional) Текст генерируется путем обработки кода виртуальной клавиши
                                            с раскладкой клавиатуры. Не требуется для событий keyUp и rawKeyDown
                                            (по умолчанию: "").
        :param unmodifiedText:          (optional) Текст, который был бы сгенерирован клавиатурой, если бы не
                                            были нажаты модификаторы (кроме Shift). Полезно для обработки
                                            сочетаний клавиш (по умолчанию: "").
        :param keyIdentifier:           (optional) Уникальный идентификатор кнопки (например, 'U+0041')
                                            (по умолчанию: "").
        :param code:                    (optional) Уникальное значение строки, определяемое DOM для каждого
                                            физического ключа (например, 'KeyA') (по умолчанию: "").
        :param key:                     (optional) Уникальное строковое значение, определяемое DOM, описывающее
                                            значение клавиши в контексте активных модификаторов, раскладки
                                            клавиатуры и т. д. (Например, 'AltGr') (по умолчанию: "").
        :param windowsVirtualKeyCode:   (optional) Код виртуальной клавиши Windows (по умолчанию: 0).
        :param nativeVirtualKeyCode:    (optional) Собственный код виртуальной клавиши (по умолчанию: 0).
        :param autoRepeat:              (optional) Было ли событие сгенерировано из автоматического повтора
                                            (по умолчанию: false).
        :param isKeypad:                (optional) Было ли событие сгенерировано с клавиатуры (по умолчанию: false).
        :param isSystemKey:             (optional) Было ли событие, событием системной кнопки (по умолчанию: false).
        :param location:                (optional) Было ли событие с левой или правой стороны клавиатуры.
                                            1 = слева, 2 = справа (по умолчанию: 0).
        :param commands:                (optional) Команды редактирования для отправки с ключевым событием (например,
                                            'selectAll') (по умолчанию: []). Они связаны с именами команд,
                                            используемыми в document.execCommand и NSStandardKeyBindingResponding,
                                            но не равны им.(по умолчанию: []).
        :return:        None
        """
        args = {
            "type": type_, "modifiers": modifiers, "text": text, "unmodifiedText": unmodifiedText,
            "keyIdentifier": keyIdentifier, "code": code, "key": key,
            "windowsVirtualKeyCode": windowsVirtualKeyCode, "nativeVirtualKeyCode": nativeVirtualKeyCode,
            "autoRepeat": autoRepeat, "isKeypad": isKeypad, "isSystemKey": isSystemKey, "location": location,
            "commands": (commands if commands else [])
        }
        if timestamp is not None:
            args.update({"timestamp": timestamp})
        else:
            args.update({"timestamp": int(time.time() * 1000)})
        await self.page_instance.Call("Input.dispatchKeyEvent", args)

    async def DispatchMouseEvent(
        self, type_: str, x: float, y: float,
        modifiers:            int = 0,
        timestamp:            Optional[int] = None,
        button:               str = "none",
        buttons:              int = 0,
        clickCount:           int = 1,
        force:              float = 0,
        tangentialPressure: float = 0,
        tiltX:                int = 0,
        tiltY:                int = 0,
        twist:                int = 0,
        deltaX:             float = 0,
        deltaY:             float = 0,
        pointerType:          str = "mouse"
    ) -> None:
        """
        Отправляет событие мышки на страницу.
        https://chromedevtools.github.io/devtools-protocol/tot/Input#method-dispatchMouseEvent
        :param type_:                   Тип события. Допустимые значения:
                                            mousePressed, mouseReleased, mouseMoved, mouseWheel
        :param x:                       Координата X события относительно области просмотра основного
                                            фрейма в пикселях CSS.
        :param y:                       Координата Y события относительно области просмотра основного
                                            фрейма в пикселях CSS.
        :param modifiers:               (optional) Битовое поле, представляющее нажатые клавиши-модификаторы.
                                            Alt = 1, Ctrl = 2, Meta/Command = 4, Shift = 8 (по умолчанию: 0).
        :param timestamp:               (optional) UNIX-time (число секунд с 1 января 1970)
        :param button:                  (optional) Кнопка мыши. Допустимые значения:
                                            none, left, middle, right, back, forward
        :param buttons:                 (optional) Число, указывающее, какие кнопки нажимаются на мыши, когда
                                            вызывается событие мыши.
                                            Left=1, Right=2, Middle=4, Back=8, Forward=16, None=0.
        :param clickCount:              (optional) Количество нажатий кнопки мыши (по умолчанию: 0).

        ========================== опции стилуса ========================
        :param force:                   (optional, EXPERIMENTAL) - Нормализованное давление(на кнопку?) в
                                            интервале от 0 до 1.
        :param tangentialPressure:      (optional, EXPERIMENTAL) - Нормализованное тангенциальное давление
                                            (на кнопку?) в интервале от -1 до 1.
        :param tiltX:                   (optional, EXPERIMENTAL) - Плоский угол между плоскостью YZ и плоскостью,
                                            содержащей ось пера и ось Y, в градусах диапазона [-90,90], положительный
                                            tiltX направлен вправо (по умолчанию: 0).
        :param tiltY:                   (optional, EXPERIMENTAL) - Плоский угол между плоскостью XZ и плоскостью,
                                            содержащей ось пера и ось X, в градусах диапазона [-90,90], положительный
                                            tiltY направлен к пользователю (по умолчанию: 0).
        :param twist:                   (optional, EXPERIMENTAL) - Вращение стилуса пера по часовой стрелке вокруг
                                            своей главной оси в градусах в диапазоне [0,359] (по умолчанию: 0).
        ========================== конец опций стилуса ========================

        :param deltaX:                  (optional) Дельта X в пикселях CSS для события колеса мыши
                                            (по умолчанию: 0).
        :param deltaY:                  (optional) Дельта Y в пикселях CSS для события колеса мыши
                                            (по умолчанию: 0).
        :param pointerType:             (optional) Тип указателя (по умолчанию: "mouse"). Допустимые значения:
                                            mouse, pen
        :return:
        """
        args = {
            "type": type_, "x": x, "y": y, "modifiers": modifiers, "button": button, "buttons": buttons,
            "clickCount": clickCount, "force": force, "tangentialPressure": tangentialPressure, "tiltX": tiltX,
            "tiltY": tiltY, "twist": twist, "deltaX": deltaX, "deltaY": deltaY, "pointerType": pointerType
        }
        if timestamp is not None:
            args.update({"timestamp": timestamp})
        else:
            args.update({"timestamp": int(time.time() * 1000)})
        await self.page_instance.Call("Input.dispatchMouseEvent", args)

    async def DispatchTouchEvent(
        self, type_: str,
        touchPoints: Optional[List[TouchPoint]] = None,
        modifiers:                int = 0,
        timestamp:                Optional[int] = None
    ) -> None:
        """
        Отправляет событие прикосновения на страницу.
        https://chromedevtools.github.io/devtools-protocol/tot/Input#method-dispatchTouchEvent
        :param type_:                   Тип сенсорного события. TouchEnd и TouchCancel не должны содержать
                                            никаких точек касания, в то время как TouchStart и TouchMove
                                            должны содержать хотя бы одну. Допустимые значения:
                                            touchStart, touchEnd, touchMove, touchCancel
        :param touchPoints:             Список точек касания на сенсорном устройстве. Генерируется одно
                                            событие на каждую измененную точку (по сравнению с предыдущим
                                            событием касания в последовательности), эмулируя
                                            нажатия/перемещения/отпускания точек одну за другой.
        :param modifiers:               (optional) Битовое поле, представляющее нажатые клавиши-модификаторы.
                                            Alt = 1, Ctrl = 2, Meta/Command = 4, Shift = 8 (по умолчанию: 0).
        :param timestamp:               (optional) UNIX-time (число секунд с 1 января 1970)
        :return:
        """
        args = { "type": type_, "modifiers": modifiers }
        if type_ == "touchEnd" or type_ == "touchCancel":
            if len(touchPoints) == 0 or touchPoints is None:
                raise ValueError(f"Для действия '{type_}', аргумент 'touchPoints' должен получить хотя-бы одну координату")
            args.update({"touchPoints": [point.to_dict() for point in touchPoints]})

        if timestamp is not None: args.update({"timestamp": timestamp})
        else: args.update({"timestamp": int(time.time() * 1000)})

        await self.page_instance.Call("Input.dispatchTouchEvent", args)

    async def InsertText(self, text: str) -> None:
        """
        (EXPERIMENTAL)
        Этот метод эмулирует вставку текста, который не поступает от нажатия клавиши, например,
            клавиатуры Emoji или IME.
        https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-insertText
        :param text:                    Текст для вставки.
        :return:
        """
        await self.page_instance.Call("Input.insertText", {"text": text})

    async def SynthesizePinchGesture(
            self, x: float, y: float, scaleFactor: float,
                relativeSpeed: Optional[int] = None,
            gestureSourceType: Optional[Literal["touch", "mouse"]] = None
    ) -> None:
        """
        (EXPERIMENTAL)
        Синтезирует жест щипка за период времени, генерируя соответствующие сенсорные события.
        https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-synthesizePinchGesture
        :param x:                       Координата X начала жеста в пикселях CSS.
        :param y:                       Координата Y начала жеста в пикселях CSS.
        :param scaleFactor:             Относительный масштабный коэффициент после увеличения
                                            (> 1,0 увеличивает, <1,0 уменьшает).
        :param relativeSpeed:           (optional) Относительная скорость указателя в пикселях в секунду (по умолчанию: 800).
        :param gestureSourceType:       (optional) Тип входных событий, который должен быть сгенерирован
                                            (по умолчанию: "default", который запрашивает у платформы
                                            предпочтительный тип ввода). default, touch, mouse
        :return:
        """
        args = {"x": x, "y": y, "scaleFactor": scaleFactor}
        if relativeSpeed is not None:
            args.update({"relativeSpeed": relativeSpeed})
        if gestureSourceType is not None:
            args.update({"gestureSourceType": gestureSourceType})
        await self.page_instance.Call("Input.synthesizePinchGesture", args)

    async def SynthesizeScrollGesture(
            self, x: float, y: float,
                      xDistance: Optional[float] = None,
                      yDistance: Optional[float] = None,
                    xOverscroll: Optional[float] = None,
                    yOverscroll: Optional[float] = None,
                    preventFling: Optional[bool] = None,
                            speed: Optional[int] = None,
                gestureSourceType: Optional[Literal["touch", "mouse"]] = None,
                      repeatCount: Optional[int] = None,
                    repeatDelayMs: Optional[int] = None,
            interactionMarkerName: Optional[str] = None
    ) -> None:
        """
        (EXPERIMENTAL)
        Синтезирует жест прокрутки в течение определенного периода времени, генерируя соответствующие сенсорные события.
            Возвращает управление только после выполнения жеста!
        https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-synthesizeScrollGesture
        :param x:                       Координата X начала жеста в пикселях CSS.
        :param y:                       Координата Y начала жеста в пикселях CSS.
        :param xDistance:               (optional) Расстояние для прокрутки вдоль оси X (положительное для прокрутки влево).
        :param yDistance:               (optional) Расстояние для прокрутки вдоль оси Y (положительное для прокрутки вверх).
        :param xOverscroll:             (optional) Количество дополнительных пикселей для прокрутки назад вдоль оси X, в
                                            дополнение к заданному расстоянию.
        :param yOverscroll:             (optional) Количество дополнительных пикселей для прокрутки назад вдоль оси Y, в
                                            дополнение к заданному расстоянию.
        :param preventFling:            (optional) Запретить сброс (по умолчанию: true).
        :param speed:                   (optional) Скорость прокрутки в пикселях в секунду (по умолчанию: 800).
        :param gestureSourceType:       (optional) Тип входных событий, который должен быть сгенерирован
                                            (по умолчанию: "default", который запрашивает у платформы
                                            предпочтительный тип ввода). default, touch, mouse
        :param repeatCount:             (optional) Количество повторений жеста (по умолчанию: 0).
        :param repeatDelayMs:           (optional) Задержка в миллисекундах между повторениями. (по умолчанию: 250).
        :param interactionMarkerName:   (optional) Имя генерируемых маркеров взаимодействия, если оно не пустое (по умолчанию: "").
        :return:
        """
        args = {"x": x, "y": y}
        if xDistance is not None:
            args.update({"xDistance": xDistance})
        if yDistance is not None:
            args.update({"yDistance": yDistance})
        if xOverscroll is not None:
            args.update({"xOverscroll": xOverscroll})
        if yOverscroll is not None:
            args.update({"yOverscroll": yOverscroll})
        if preventFling is not None:
            args.update({"preventFling": preventFling})
        if speed is not None:
            args.update({"speed": speed})
        if gestureSourceType is not None:
            args.update({"gestureSourceType": gestureSourceType})
        if repeatCount is not None:
            args.update({"repeatCount": repeatCount})
        if repeatDelayMs is not None:
            args.update({"repeatDelayMs": repeatDelayMs})
        if interactionMarkerName is not None:
            args.update({"interactionMarkerName": interactionMarkerName})
        await self.page_instance.Call("Input.synthesizeScrollGesture", args)

    async def SynthesizeTapGesture(
            self, x: float, y: float,
            duration:          Optional[int] = None,
            tapCount:          Optional[int] = None,
            gestureSourceType: Optional[Literal["touch", "mouse"]] = None
    ) -> None:
        """
        (EXPERIMENTAL)
        Синтезирует жест касания за период времени, генерируя соответствующие сенсорные события.
        https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-synthesizeTapGesture
        :param x:                       Координата X места касания в пикселях CSS.
        :param y:                       Координата Y места касания в пикселях CSS.
        :param duration:                (optional) Продолжительность касания (по умолчанию: 50).
        :param tapCount:                (optional) Количество прикосновений (по умолчанию: 1).
        :param gestureSourceType:       (optional) Тип входных событий, который должен быть сгенерирован
                                            (по умолчанию: "default", который запрашивает у платформы
                                            предпочтительный тип ввода). default, touch, mouse
        :return:
        """
        args = {"x": x, "y": y}
        if duration is not None:
            args.update({"duration": duration})
        if tapCount is not None:
            args.update({"tapCount": tapCount})
        if gestureSourceType is not None:
            args.update({"gestureSourceType": gestureSourceType})
        await self.page_instance.Call("Input.synthesizeTapGesture", args)

    async def ActivateTarget(self, targetId: Optional[str] = None) -> None:
        """
        Активирует (создаёт фокус) "target".
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-activateTarget
        :param targetId:        Строка, представляющая идентификатор созданной страницы.
        :return:
        """
        if targetId is None: targetId = self.page_instance.page_id
        await self.page_instance.Call("Target.activateTarget", {"targetId": targetId})

    async def BringToFront(self) -> None:
        """
        Выводит страницу на передний план (активирует вкладку).
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-bringToFront
        :return:
        """
        await self.page_instance.Call("Page.bringToFront")

    # endregion

    # ==================================================================================================================

    async def SwipeTo(
        self,
        direction: Literal["up", "down", "left", "right"],
        x:            float = 0,
        y:            float = 0,
        distance:     Optional[float] = None,
        speed:        Optional[float] = None,
        overscroll:   Optional[float] = None,
        repeat_count:   Optional[int] = None,
        repeat_delay: Optional[float] = None,
        gestureSourceType: Literal["touch", "mouse"] = "mouse"
    ) -> None:
        """
        Скроллит вьюпорт жестом "touch" на всю его длину/высоту.
            Возвращает управление только после выполнения жеста!
        :param direction:           (optional) Направление. Может быть следующим:
                                        up — пальцем вверх(прокрутка вниз)
                                        down — пальцем вниз(прокрутка вверх)
                                        left — пальцем влево(прокрутка вправо)
                                        right — пальцем вправо(прокрутка влево)
        :param x:                   (optional) X-координата начальной точки.
        :param y:                   (optional) Y-координата начальной точки.
        :param distance:            (optional) Дистанция движения.
        :param speed:               (optional) Скорость движения(по умолчанию = 800).
        :param overscroll:          (optional) Дополнительная дистанция в пикселях.
        :param repeat_count:        (optional) Кол-во повторений сделанного жеста.
        :param repeat_delay:        (optional) Задержка между повторениями.
        :param gestureSourceType:   (optional) Задержка между повторениями.
        :return:
        """
        if direction not in ["up", "down", "left", "right"]:
            raise ValueError("'direction' must be one from — up; down; left; right")
        sign = -1 if direction in ["up", "left"] else 1
        rect = None
        if x is None:
            rect = await self.page_instance.GetViewportRect()
            x = 10 if direction == "right" else rect.width - 10 if direction == "left" else rect.height / 2

        if y is None:
            rect = await self.page_instance.GetViewportRect()
            y = 10 if direction == "down" else rect.height - 10 if direction == "up" else rect.width / 2

        if distance is None:
            rect = rect or await self.page_instance.GetViewportRect()
            distance = (rect.height if direction in ["up", "down"] else rect.width) - 10

        overscroll = overscroll if overscroll is not None else 0

        args = {
            "x": x, "y": y, "speed": speed, "repeatCount": repeat_count,
            "repeatDelayMs": repeat_delay, "gestureSourceType": gestureSourceType,
            "xDistance": distance * sign if direction in ["left", "right"] else None,
            "yDistance": distance * sign if direction in ["up", "down"] else None,
            "xOverscroll": overscroll * -sign if direction in ["left", "right"] else None,
            "yOverscroll": overscroll * -sign if direction in ["up", "down"] else None,
        }
        await self.SynthesizeScrollGesture(**args)

    async def ClickTo(self, x: int, y: int, delay: float = None) -> None:
        """
        Эмулирует клик мыши по координатам.
        :param x:               x - координата
        :param y:               y - координата
        :param delay:           задержка перед отпусканием
        :return:
        """
        await self.DispatchMouseEvent("mousePressed", x, y, button="left")
        if delay: await asyncio.sleep(delay)
        await self.DispatchMouseEvent("mouseReleased", x, y, button="left")

    async def MouseMoveTo(self, x: int, y: int) -> None:
        await self.DispatchMouseEvent("mouseMoved", x, y)

    async def MouseWheel(self, x: float, y: float, deltaX: float = 0, deltaY: float = 0) -> None:
        """
        Крутит колесо мышки.
        :param x:               Положение
        :param y:               указателя мыши.
        :param deltaX:          Скроллит по-горизонтали.
        :param deltaY:          Скроллит по-вертикали, положительное значение == вниз.
        :return:
        """
        await self.DispatchMouseEvent("mouseWheel", x, y, button="middle", deltaX=deltaX, deltaY=deltaY)

    async def WheelTo(self, direction: str = "down") -> None:
        sign = -1 if direction in ["up", "left"] else 1
        rect = await self.page_instance.GetViewportRect()
        x = 10 if direction == "right" else rect.width - 10 if direction == "left" else rect.height / 2
        y = 10 if direction == "down" else rect.height - 10 if direction == "up" else rect.width / 2
        distance = ((rect.height if direction in ["up", "down"] else rect.width) - 10) * sign
        delta = {"deltaX": distance} if direction in ["left", "rigth"] else {"deltaY": distance}
        await self.MouseWheel(x, y, **delta)


    async def MouseMoveToCoordinatesAndClick(self, x: int, y: int) -> None:
        await self.MouseMoveTo(x, y)
        await self.ClickTo(x, y)

    async def SendChar(self, char: str) -> None:
        """
        Эмулирует ввод символа нажатием соответствующей кнопки клавиатуры.
            Внимание! Курсор должен быть установлен в эдит-боксе!
        :param char:             Символ для ввода.
        :return:
        """
        upper_key = char.upper()
        args = {
            "text": char, "key": char, "keyIdentifier": f"U+{WINDOWS_KEY_SET[upper_key]:X}",
            "windowsVirtualKeyCode": WINDOWS_KEY_SET[upper_key],
            "nativeVirtualKeyCode": WINDOWS_KEY_SET[upper_key]
        }
        if len(char) > 1: raise ValueError(f"Передаваемая строка: '{char}' — должна быть из одного символа!")
        await self.DispatchKeyEvent("char", **args)

    async def SendText(
            self, text: str, interval: Optional[Tuple[float, float]] = None
    ) -> None:
        """
        Эмулирует последовательный набор текста.
            Внимание! Курсор должен быть установлен в эдит-боксе!
        :param text:             Последовательность символов для ввода.
        :param interval:         Задержка после нажатия кнопки. None - отключает задержку.
                                     Кортеж из двух чисел описывает интервал, вычисляемый рандомно.
                                     Кортеж (10, None) — устанавливает фиксированное ожидание в 10 секунд
        :return:
        """
        for letter in text:
            await self.SendChar(letter)
            if interval is not None: await asyncio.sleep(random.uniform(interval[0], interval[1]))

    async def SendKeyEvent(self, event: dict, *modifiers: KeyModifiers) -> None:
        args = {}
        if modifiers:
            s = 0
            for m in modifiers: s += m.value
            args.update({"modifiers": s})

        args.update(event)
        await self.DispatchKeyEvent("keyDown", **args)
        await self.DispatchKeyEvent("keyUp", **args)

    async def ControlA(self) -> None:
        """ Выделить весь текст(Ctrl+A). """
        await self.SendKeyEvent(KeyEvents.keyA, KeyModifiers.ctrl)

    async def BackspaceText(
            self, count: int = 1, modifier: KeyModifiers = KeyModifiers.none) -> None:
        """
        Удаляет текст в текстовом поле с позиции курсора по направлению в лево, или полностью очистить,
            эмулируя нажатие клавиши 'Backspace'.
            Внимание! Курсор должен быть установлен в эдит-боксе!
        :param count:                   (optional) Количество нажатий. Не имеет воздействия при полной очистке.
        :param modifier:                (optional) none - удалить один символ, alt - слово, включая стоящие перед
                                            ним пробелы, ctrl — полностью очистить эдит-бокс.
        :return:
        """

        if modifier.name == "ctrl":
            await self.ControlA()
            await self.SendKeyEvent(KeyEvents.backspace)
            return

        for i in range(count):
            await self.SendKeyEvent(KeyEvents.backspace, modifier.value)

    async def SetWindowBounds(self, bounds: WindowBounds, windowId: Optional[int] = None) -> None:
        """
        (EXPERIMENTAL)
        Устанавливает позицию и/или размер окна.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-setWindowBounds
        :param bounds:          Новые границы окна, а так же состояние.
        :param windowId:        Идентификатор окна.

        :return:        None
        """
        if windowId is None:
            windowId = (await self.page_instance.GetWindowForTarget()).windowId
        await self.page_instance.Call("Browser.setWindowBounds", {"windowId": windowId, "bounds": bounds.to_dict()})



from abc import ABC, abstractmethod
from typing import Optional, Union, List, Callable, Awaitable
from ..Data import DomainEvent
from ..domains.Network import NetworkType
from dataclasses import dataclass, field


class Fetch(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Fetch
    """
    __slots__ = ()

    def __init__(self):
        self.fetch_domain_enabled = False

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def FetchEnable(
        self,
        patterns: Optional[List["FetchType.RequestPattern"]] = None,
        handleAuthRequests: bool = False,
        on_pause: Optional[Callable[['FetchType.EventRequestPaused'], Awaitable[None]]] = None,
        on_auth:  Optional[Callable[['FetchType.EventAuthRequired'], Awaitable[None]]] = None,
    ) -> None:
        """
        Включает выдачу событий requestPaused. Запрос будет приостановлен до тех пор,
            пока клиент не вызовет одну из функций failRequest, fulfillRequest или
            continueRequest / continueWithAuth.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-enable
        :param patterns:            (optional) Если указано, только запросы,
                                        соответствующие любому из этих шаблонов, будут
                                        вызывать событие fetchRequested и будут
                                        приостановлены до ответа клиента. Если не
                                        установлен, все запросы будут затронуты.
                                            https://chromedevtools.github.io/devtools-protocol/tot/Fetch#type-RequestPattern
        :param handleAuthRequests:  (optional) Если True - события authRequired будут
                                        выдаваться и запросы будут приостановлены в
                                        ожидании вызова continueWithAuth.
        :param on_pause:            (optional) Корутина, которая будет получать все события "requestPaused".
        :param on_auth:             (optional) Корутина, которая будет получать все события "authRequired".
        :return:
        """
        async def on_pause_decorator(
                params: dict, func: Callable[["FetchType.EventRequestPaused"], Awaitable[None]]) -> None:
            await func(FetchType.EventRequestPaused(**params))

        async def on_auth_decorator(
                params: dict, func: Callable[["FetchType.EventAuthRequired"], Awaitable[None]]) -> None:
            await func(FetchType.EventAuthRequired(**params))

        if on_pause is not None: await self.AddListenerForEvent(FetchEvent.requestPaused, on_pause_decorator, on_pause)
        if on_auth is not None: await self.AddListenerForEvent(FetchEvent.authRequired, on_auth_decorator, on_auth)

        args = {}
        patterns = patterns if patterns is not None else []
        if patterns:
            args.update({"patterns": [p.dict() for p in patterns]})
        if handleAuthRequests: args.update({"handleAuthRequests": handleAuthRequests})
        await self.Call("Fetch.enable", args)
        self.fetch_domain_enabled = True

    async def FetchDisable(self) -> None:
        """
        Отключает взаимодействие с доменом.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-disable
        :return:
        """
        await self.Call("Fetch.disable")
        self.fetch_domain_enabled = False

    async def FailRequest(self, requestId: str, errorReason: str) -> None:
        """
        Вызывает сбой запроса по указанной причине.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-failRequest
        :param requestId:           Идентификатор, полученный клиентом в событии requestPaused.
        :param errorReason:         Причина сбоя выборки на уровне сети. Возможные значения:
                                        Failed, Aborted, TimedOut, AccessDenied,
                                        ConnectionClosed, ConnectionReset,
                                        ConnectionRefused, ConnectionAborted,
                                        ConnectionFailed, NameNotResolved,
                                        InternetDisconnected, AddressUnreachable,
                                        BlockedByClient, BlockedByResponse
        :return:
        """
        await self.Call("Fetch.failRequest", {"requestId": requestId, "errorReason": errorReason})

    async def FulfillRequest(
            self, requestId: str,
            responseCode:                     int = 200,
            responseHeaders: Optional[List[dict]] = None,
            binaryResponseHeaders:  Optional[str] = None,
            body:                   Optional[str] = None,
            responsePhrase:         Optional[str] = None,
            wait_for_response:               bool = False
    ) -> None:
        """
        Предоставляет ответ на запрос.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-fulfillRequest
        :param requestId:               Идентификатор, полученный клиентом в событии requestPaused.
        :param responseCode:            Код ответа HTTP(например - 200).
        :param responseHeaders:         (optional) Заголовки ответа. Например:
                                            [
                                                { "name": "User-Agent", "value": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.149 Safari/537.36" },
                                                { "name": "Content-Type", "value": "application/json; charset=UTF-8" }
                                            ]
        :param binaryResponseHeaders:   (optional) Альтернативный способ указания заголовков ответа в
                                            виде разделенных \0 серий пар имя-значение. Описанный выше
                                            метод предпочтительней, если вам не нужно представлять
                                            некоторые значения, отличные от UTF8, которые не могут быть
                                            переданы по протоколу, в виде текста.
        :param body:                    (optional) Тело ответа, кодированное в строку формата base64.
        :param responsePhrase:          (optional) Текстовое представление responseCode. Если
                                            отсутствует, используется стандартная фраза,
                                            соответствующая responseCode.
        :param wait_for_response:       (optional) Дожидаться ответа?
        :return:
        """
        args = {"requestId": requestId, "responseCode": responseCode}
        if responseHeaders is not None: args.update({"responseHeaders": responseHeaders})
        if binaryResponseHeaders is not None: args.update({"binaryResponseHeaders": binaryResponseHeaders})
        if body is not None: args.update({"body": body})
        if responsePhrase is not None: args.update({"responsePhrase": responsePhrase})
        # print("fulfillRequest args", json.dumps(args, indent=4))
        await self.Call("Fetch.fulfillRequest", args, wait_for_response=wait_for_response)

    async def ContinueRequest(
        self, requestId: str,
        url:                Optional[str] = None,
        method:             Optional[str] = None,
        postData:           Optional[str] = None,
        headers:     Optional[List[dict]] = None,
        wait_for_response:           bool = False
    ) -> None:
        """
        Продолжает запрос, дополнительно изменяя некоторые его параметры.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-continueRequest
        :param requestId:           Идентификатор, полученный клиентом в событии requestPaused.
        :param url:                 (optional) Если установлено, URL-адрес запроса будет изменен так,
                                        чтобы страница не наблюдалась.
        :param method:              (optional) Переопределяет метод запроса переданным значением.
        :param postData:            (optional) Переопределяет данные запроса переданными.
        :param headers:             (optional) Переопределяет заголовки запроса переданными. . Например:
                                        [
                                            { "name": "User-Agent", "value": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.149 Safari/537.36" },
                                            { "name": "Content-Type", "value": "application/json; charset=UTF-8" }
                                        ]
        :param wait_for_response:       (optional) Дожидаться ответа?
        :return:
        """
        args = {"requestId": requestId}
        if url is not None: args.update({"url": url})
        if method is not None: args.update({"method": method})
        if postData is not None: args.update({"postData": postData})
        if headers is not None: args.update({"headers": headers})
        await self.Call("Fetch.continueRequest", args, wait_for_response=wait_for_response)

    async def ContinueWithAuth(self, requestId: str, authChallengeResponse: list) -> None:
        """
        Продолжает запрос, предоставляющий authChallengeResponse после события authRequired.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-continueWithAuth
        :param requestId:               Идентификатор, полученный клиентом в событии requestPaused.
        :param authChallengeResponse:   Ответ с помощью authChallenge.
                                            {
                                                "response": str(), -> Решение о том, что делать в ответ
                                                    на запрос авторизации. По умолчанию означает
                                                    использование стандартного поведения сетевого стека,
                                                    что, скорее всего, приведет к отмене проверки
                                                    подлинности или появлению всплывающего диалогового
                                                    окна. Возможные значения:
                                                        Default, CancelAuth, ProvideCredentials
                                                "username": str(), -> (optional) Имя пользователя для
                                                    предоставления, может быть пустым. Устанавливается,
                                                    только если ответом является ProvideCredentials.
                                                "password": str(), -> (optional) Пароль пользователя для
                                                    предоставления, может быть пустым. Устанавливается,
                                                    только если ответом является ProvideCredentials.
                                            }
        :return:
        """
        args = {"requestId": requestId, "authChallengeResponse": authChallengeResponse}
        await self.Call("Fetch.continueWithAuth", args)

    async def GetResponseBody(self, requestId: str) -> dict:
        """
        Вызывает тело ответа, получаемого от сервера и возвращаемого в виде одной строки. Может
            выдаваться только для запроса, который приостановлен на этапе ответа и является
            взаимоисключающим с "takeResponseBodyForInterceptionAsStream". Вызов других методов,
            влияющих на запрос, или отключение домена выборки до получения тела приводит к
            неопределенному поведению.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-getResponseBody
        :param requestId:               Идентификатор перехваченного запроса для получения его тела.
        :return:                        {
                                            "body": str(),        -> Тело ответа.
                                            "base64Encoded": bool -> True - если контент кодирован
                                                                        как base64.
                                        }
        """
        return await self.Call("Fetch.getResponseBody", {"requestId": requestId})

    async def TakeResponseBodyAsStream(self, requestId: str) -> dict:
        """
        Возвращает дескриптор потока, представляющего тело ответа. Запрос должен быть приостановлен
            на этапе HeadersReceived. Обратите внимание, что после этой команды запрос не может быть
            продолжен как есть - клиент должен либо отменить его, либо предоставить тело ответа.
            Поток поддерживает только последовательное чтение, IO.read потерпит неудачу, если указана
            позиция. Этот метод является взаимоисключающим с getResponseBody. Вызов других методов,
            влияющих на запрос, или отключение домена выборки до получения тела приводит к
            неопределенному поведению.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-takeResponseBodyAsStream
        :param requestId:               Идентификатор перехваченного запроса для получения его тела.
        :return:                        {
                                            "stream": str(), -> Это либо получается из другого метода,
                                                либо указывается как blob <uuid> это UUID Blob.
                                                    IO.StreamHandle:
                                                    https://chromedevtools.github.io/devtools-protocol/tot/IO#type-StreamHandle
                                        }
        """
        return await self.Call("Fetch.takeResponseBodyAsStream", {"requestId": requestId})

    @abstractmethod
    async def Call(
        self, domain_and_method: str,
        params: Optional[dict] = None,
        wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")

    @abstractmethod
    async def AddListenerForEvent(
            self, event: Union[str, DomainEvent], listener: Callable, *args: any) -> None:
        raise NotImplementedError("async method AddListenerForEvent() — is not implemented")

    @abstractmethod
    def RemoveListenersForEvent(self, event: str) -> None:
        raise NotImplementedError("method RemoveListenersForEvent() — is not implemented")


class FetchEvent(DomainEvent):
    authRequired = "Fetch.authRequired"
    requestPaused = "Fetch.requestPaused"


class FetchType:

    @dataclass
    class EventRequestPaused:
        requestId: str
        frameId: str
        resourceType: str                               # ! Allowed Values: Document, Stylesheet, Image, Media, Font,
                                                        # !     Script, TextTrack, XHR, Fetch, EventSource, WebSocket,
                                                        # !     Manifest, SignedExchange, Ping, CSPViolationReport,
                                                        # !     Preflight, Other
        request: NetworkType.Request
        responseHeaders: list["FetchType.HeaderEntry"]
        responseErrorReason: Optional[str] = None       # ! Allowed Values: Failed, Aborted, TimedOut, AccessDenied,
                                                        # !     ConnectionClosed, ConnectionReset, ConnectionRefused,
                                                        # !     ConnectionAborted, ConnectionFailed, NameNotResolved,
                                                        # !     InternetDisconnected, AddressUnreachable,
                                                        # !     BlockedByClient, BlockedByResponse
        responseStatusCode: Optional[int] = None
        responseStatusText: Optional[str] = None
        networkId: Optional[str] = None                 # ! id of request
        _request: NetworkType.Request = field(init=False, repr=False, default=None)
        _responseHeaders: Optional[list["FetchType.HeaderEntry"]] = field(init=False, repr=False, default=None)

        @property
        def request(self) -> NetworkType.Request:
            return self._request

        @request.setter
        def request(self, data: dict) -> None:
            self._request = NetworkType.Request(**data)

        @property
        def responseHeaders(self) -> Optional[list["FetchType.HeaderEntry"]]:
            return self._responseHeaders

        @responseHeaders.setter
        def responseHeaders(self, data: list[dict[str, str]]) -> None:
            if not isinstance(data, property):
                self._responseHeaders = [FetchType.HeaderEntry(**item) for item in data]
            else:
                self._responseHeaders = None


    @dataclass
    class EventAuthRequired:
        requestId: str
        request: NetworkType.Request
        frameId: str
        resourceType: str                               # ! Allowed Values: Document, Stylesheet, Image, Media, Font,
                                                        # !     Script, TextTrack, XHR, Fetch, EventSource, WebSocket,
                                                        # !     Manifest, SignedExchange, Ping, CSPViolationReport,
                                                        # !     Preflight, Other
        authChallenge: "FetchType.AuthChallenge"
        _authChallenge: "FetchType.AuthChallenge" = field(init=False, repr=False, default=None)

        @property
        def authChallenge(self) -> "FetchType.AuthChallenge":
            return self._authChallenge

        @authChallenge.setter
        def authChallenge(self, data: dict) -> None:
            self._authChallenge = FetchType.AuthChallenge(**data)


    @dataclass
    class AuthChallenge:
        origin: str
        scheme: str
        realm: str                                      # ! May be empty.
        source: Optional[str] = None


    @dataclass
    class HeaderEntry:
        name: str
        value: str


    @dataclass
    class RequestPattern:
        urlPattern: str = "*"                           # ? Подстановочные знаки ( '*'-> ноль или более, '?'-> ровно
                                                        # ?     один) допускаются. Экранирующий символ — обратная
                                                        # ?     косая черта(\).
        resourceType: Optional[str] = None              # ? Если установлено, будут перехватываться только соответствующие
                                                        # ?     указанным типам.
                                                        # ! Allowed Values: Document, Stylesheet, Image, Media, Font,
                                                        # !     Script, TextTrack, XHR, Fetch, EventSource, WebSocket,
                                                        # !     Manifest, SignedExchange, Ping, CSPViolationReport,
                                                        # !     Preflight, Other
        requestStage: Optional[str] = None              # ? Стадия перехвата запроса. Запрос будет перехвачен до того,
                                                        # ?     как запрос будет отправлен. Ответ будет перехвачен после
                                                        # ?     получения ответа (но до получения тела ответа).
                                                        # ! Allowed Values: Request, Response

        def dict(self) -> dict:
            result = {}
            for k, v in self.__dict__.items():
                if v: result[k] = v
            return result


try:
    import ujson as json
except ModuleNotFoundError:
    import json
import asyncio
import websockets.client

from .exceptions import EvaluateError, get_cdtp_error, highlight_eval_error, PromiseEvaluateError, \
    highlight_promise_error
from .utils import log

from websockets.exceptions import ConnectionClosedError
from inspect import iscoroutinefunction
from typing import Callable, Awaitable, Optional, Union, Tuple, List, Dict
from abc import ABC
from .Data import DomainEvent, Sender, Receiver, CommonCallback


class AbsPage(ABC):
    __slots__ = (
        "ws_url", "page_id", "frontend_url", "callback", "is_headless_mode", "verbose", "browser_name", "id",
        "responses", "connected", "ws_session", "receiver", "on_detach_listener", "listeners", "listeners_for_event",
        "runtime_enabled", "on_close_event", "context_manager", "_connected", "_page_id", "_verbose",
        "_browser_name", "_is_headless_mode"
    )

    def __init__(
        self,
        ws_url:           str,
        page_id:          str,
        frontend_url:     str,
        callback:         CommonCallback,
        is_headless_mode: bool,
        verbose:          bool,
        browser_name:     str
    ) -> None:
        self.ws_url            = ws_url
        self.page_id           = page_id
        self.frontend_url      = frontend_url
        self.callback          = callback
        self.is_headless_mode  = is_headless_mode

        self.verbose           = verbose
        self.browser_name      = browser_name

        self.id                = 0
        self.connected         = False
        self.ws_session        = None
        self.receiver          = None
        self.on_detach_listener: List[Callable[[any], Awaitable[None]], list, dict] = []
        self.listeners         = {}
        self.listeners_for_event = {}
        self.runtime_enabled     = False
        self.on_close_event = asyncio.Event()
        self.responses: Dict[int, Optional[Sender[dict]]] = {}


class Page(AbsPage):
    """
    Инстанс страницы должен быть активирован после создания вызовом метода Activate(),
        это создаст подключение по WebSocket и запустит задачи обеспечивающие
        обратную связь. Метод GetPageBy() инстанса браузера, заботится об этом
        по умолчанию.

    Если инстанс страницы более не нужен, например, при перезаписи в него нового
        инстанса, перед этим [-!-] ОБЯЗАТЕЛЬНО [-!-] - вызовите у него метод
        Detach(), или закройте вкладку/страницу браузера, с которой он связан,
        тогда это будет выполнено автоматом. Иначе в цикле событий останутся
        задачи связанные с поддержанием соединения, которое более не востребовано.
    """
    __slots__ = ()

    def __init__(
        self,
        ws_url:           str,
        page_id:          str,
        frontend_url:     str,
        callback:         CommonCallback,
        is_headless_mode: bool,
        verbose:          bool,
        browser_name:     str
    ) -> None:
        """
        :param ws_url:              Адрес WebSocket
        :param page_id:             Идентификатор страницы
        :param frontend_url:        devtoolsFrontendUrl по которому происходит подключение к дебаггеру
        :param callback:            Колбэк, который будет получать все данные,
                                        приходящие по WebSocket в виде словарей
        :param is_headless_mode:    "Headless" включён?
        :param verbose:             Печатать некие подробности процесса?
        :param browser_name:        Имя браузера
        """
        AbsPage.__init__(self, ws_url, page_id, frontend_url, callback, is_headless_mode, verbose, browser_name)

    @property
    def connected(self) -> bool: return self._connected

    @connected.setter
    def connected(self, value) -> None: self._connected = value

    @property
    def page_id(self) -> str: return self._page_id

    @page_id.setter
    def page_id(self, value) -> None: self._page_id = value

    @property
    def verbose(self) -> bool: return self._verbose

    @verbose.setter
    def verbose(self, value) -> None: self._verbose = value

    @property
    def browser_name(self) -> str: return self._browser_name

    @browser_name.setter
    def browser_name(self, value) -> None: self._browser_name = value

    @property
    def is_headless_mode(self) -> bool: return self._is_headless_mode

    @is_headless_mode.setter
    def is_headless_mode(self, value) -> None: self._is_headless_mode = value

    async def Call(
        self, domain_and_method: str,
        params:  Optional[dict] = None,
        wait_for_response: bool = True
    ) -> Optional[dict]:
        self.id += 1
        _id = self.id
        data = {
            "id": _id,
            "params": params if params else {},
            "method": domain_and_method
        }

        if not wait_for_response:
            self.responses[_id] = None
            await self._Send(json.dumps(data))
            return

        que = asyncio.Queue()
        sender, receiver = Sender[dict](que), Receiver[dict](que)
        self.responses[_id] = sender

        await self._Send(json.dumps(data))

        response = await receiver.recv()
        if "error" in response:

            if ex := get_cdtp_error(response['error']['message']):
                raise ex(f"domain_and_method = '{domain_and_method}' | params = '{str(params)}'")

            raise Exception(
                "Browser detect error:\n" +
                f"error code -> '{response['error']['code']}';\n" +
                f"error message -> '{response['error']['message']}'\n"+
                f"domain_and_method = '{domain_and_method}' | params = '{str(params)}'"
            )

        return response["result"]

    async def Eval(
        self, expression: str,
        objectGroup:            str = "console",
        includeCommandLineAPI: bool = True,
        silent:                bool = False,
        returnByValue:         bool = False,
        userGesture:           bool = True,
        awaitPromise:          bool = False
    ) -> dict:
        response = await self.Call(
            "Runtime.evaluate", {
                "expression": expression,
                "objectGroup": objectGroup,
                "includeCommandLineAPI": includeCommandLineAPI,
                "silent": silent,
                "returnByValue": returnByValue,
                "userGesture": userGesture,
                "awaitPromise": awaitPromise
            }
        )
        if "exceptionDetails" in response:
            raise EvaluateError(
                highlight_eval_error(response["result"]["description"], expression)
            )
        return response["result"]

    async def _Send(self, data: str) -> None:
        if self.connected:
            await self.ws_session.send(data)

    async def _Recv(self) -> None:
        while self.connected:
            try:
                data_msg: dict = json.loads(await self.ws_session.recv())
            # ! Браузер разорвал соединение
            except ConnectionClosedError as e:
                if self.verbose: log(f"ConnectionClosedError {e!r}")
                await self.Detach()
                return

            if ("method" in data_msg and data_msg["method"] == "Inspector.detached"
                    and data_msg["params"]["reason"] == "target_closed"):
                self.on_close_event.set()
                await self.Detach()
                return

            # Ожидающие ответов вызовы API получают ответ по id входящих сообщений.
            if sender := self.responses.pop(data_msg.get("id"), None):
                await sender.send(data_msg)

            # Если коллбэк функция была определена, она будет получать все
            #   уведомления из инстанса страницы.
            if self.callback is not None:
                asyncio.create_task(self.callback(data_msg))

            # Достаточно вызвать в контексте страницы следующее:
            # console.info(JSON.stringify({
            #     func_name: "test_func",
            #     args: [1, "test"]
            # }))
            # и если среди зарегистрированных слушателей есть с именем "test_func",
            #   то он немедленно получит распакованный список args[ ... ], вместе
            #   с переданными ему аргументами, если таковые имеются.
            if (method := data_msg.get("method")) == "Runtime.consoleAPICalled":
                # ? Был вызван домен "info"
                if data_msg["params"].get("type") == "info":

                    str_value = data_msg["params"]["args"][0].get("value")
                    try:
                        value: dict = json.loads(str_value)
                    except ValueError as e:
                        if self.verbose:
                            log(f"ValueError {e!r}")
                            log(f"Msg from browser {str_value!r}")
                        raise

                    # ? Есть ожидающие слушатели
                    if self.listeners:

                        # ? Если есть ожидающая корутина
                        if listener := self.listeners.get( value.get("func_name") ):
                            asyncio.create_task(
                                listener["function"](                               # корутина
                                    *(value["args"] if "args" in value else []),    # её список аргументов вызова
                                    *listener["args"]                               # список bind-агрументов
                                )
                            )

            # По этой же схеме будут вызваны все слушатели для обработки
            #   определённого метода, вызванного в контексте страницы,
            #   если для этого метода они зарегистрированы.
            if (    # =============================================================
                    self.listeners_for_event
                            and
                    method in self.listeners_for_event
            ):      # =============================================================
                # Получаем словарь слушателей, в котором ключи — слушатели,
                #   значения — их аргументы.
                listeners: dict = self.listeners_for_event[ method ]
                p = data_msg.get("params")
                for listener, args in listeners.items():
                    asyncio.create_task(
                        listener(                                           # корутина
                            p if p is not None else {},                     # её "params" — всегда передаётся
                            *args                                           # список bind-агрументов
                        )
                    )

    async def EvalPromise(self, script: str) -> dict:
        """ Выполняет асинхронный код на странице и возвращает результат.
        !!! ВАЖНО !!! Выполняемый код не может возвращать какие-либо JS
        типы, поэтому должен возвращать JSON-сериализованный набор данных.
        """
        result = await self.Eval(script)
        args = dict(
            promiseObjectId=result["objectId"],
            returnByValue=False,
            generatePreview=False
        )
        response = await self.Call("Runtime.awaitPromise", args)
        if "exceptionDetails" in response:
            raise PromiseEvaluateError(
                highlight_promise_error(response["result"]["description"]) +
                "\n" + json.dumps(response["exceptionDetails"])
            )
        return json.loads(response["result"]["value"])

    async def WaitForClose(self) -> None:
        """ Дожидается, пока не будет потеряно соединение со страницей. """
        await self.on_close_event.wait()

    async def Detach(self) -> None:
        """
        Отключается от инстанса страницы. Вызывается автоматически при закрытии браузера,
            или инстанса текущей страницы. Принудительный вызов не закрывает страницу,
            а лишь разрывает с ней соединение.
        """
        if not self.connected:
            return

        self.receiver.cancel()
        if self.verbose: log(f"[ DETACH ] {self.page_id}")
        self.connected = False

        if self.on_detach_listener:
            function, args, kvargs = self.on_detach_listener
            await function(*args, **kvargs)

    def RemoveOnDetach(self) -> None:
        self.on_detach_listener = []

    def SetOnDetach(self, function: Callable[[any], Awaitable[None]], *args, **kvargs) -> bool:
        """
        Регистрирует асинхронный коллбэк, который будет вызван с соответствующими аргументами
            при разрыве соединения со страницей.
        """
        if not iscoroutinefunction(function):
            raise TypeError("OnDetach-listener must be a async callable object!")
        if not self.connected: return False
        self.on_detach_listener = [function, args, kvargs]
        return True

    async def Activate(self) -> None:
        self.ws_session = await websockets.client.connect(self.ws_url, ping_interval=None)
        self.connected = True
        self.receiver = asyncio.create_task(self._Recv())
        if self.callback is not None:
            await self.Call("Runtime.enable")
            self.runtime_enabled = True

    async def AddListener(self, listener: Callable[[any], Awaitable[None]], *args: any) -> None:
        """
        Добавляет 'слушателя', который будет ожидать свой вызов по имени функции.
            Вызов слушателей из контекста страницы осуществляется за счёт
            JSON-сериализованного объекта, отправленного сообщением в консоль,
            через домен 'info'. Объект должен содержать два обязательных свойства:
                funcName — имя вызываемого слушателя
                args:    — массив аргументов

            Например, вызов javascript-кода:
                console.info(JSON.stringify({
                    funcName: "test_func",
                    args: [1, "test"]
                }))
            Вызовет следующего Python-слушателя:
                async def test_func(id, text, action):
                    print(id, text, action)
            Зарегистрированного следующим образом:
                await page.AddListener(test_func, "test-action")

            !!! ВНИМАНИЕ !!! В качестве слушателя может выступать ТОЛЬКО асинхронная
                функция, или метод.

        :param listener:        Асинхронная функция.
        :param args:            (optional) любое кол-во аргументов, которые будут переданы
                                    в функцию последними.
        :return:        None
        """
        if not iscoroutinefunction(listener):
            raise TypeError("Listener must be a async callable object!")
        if listener.__name__ not in self.listeners:
            self.listeners[ listener.__name__ ] = {"function": listener, "args": args}
            if not self.runtime_enabled:
                await self.Call("Runtime.enable")
                self.runtime_enabled = True

    async def AddListeners(
            self, *list_of_tuple_listeners_and_args: Tuple[Callable[[any], Awaitable[None]], list]) -> None:
        """
        Делает то же самое, что и AddListener(), но может зарегистрировать сразу несколько слушателей.
            Принимает список кортежей с двумя элементами, вида (async_func_or_method, list_args), где:
                async_func_or_method    - асинхронная фукция или метод
                list_args               - список её аргументов(может быть пустым)
        """
        for action in list_of_tuple_listeners_and_args:
            listener, args = action
            if not iscoroutinefunction(listener):
                raise TypeError("Listener must be a async callable object!")
            if listener.__name__ not in self.listeners:
                self.listeners[listener.__name__] = {"function": listener, "args": args}
                if not self.runtime_enabled:
                    await self.Call("Runtime.enable")
                    self.runtime_enabled = True

    def RemoveListener(self, listener: Callable[[any], Awaitable[None]]) -> None:
        """
        Удаляет слушателя.
        :param listener:        Колбэк-функция.
        :return:        None
        """
        if not iscoroutinefunction(listener):
            raise TypeError("Listener must be a async callable object!")
        if listener.__name__ in self.listeners:
            del self.listeners[ listener.__name__ ]

    async def AddListenerForEvent(
        self, event: Union[str, DomainEvent], listener: Callable[[any], Awaitable[None]], *args) -> None:
        """
        Регистирует слушателя, который будет вызываться при вызове определённых событий
            в браузере. Список таких событий можно посмотреть в разделе "Events" почти
            у каждого домена по адресу: https://chromedevtools.github.io/devtools-protocol/
            Например: 'DOM.attributeModified'
        !Внимание! Каждый такой слушатель должен иметь один обязательный 'data'-аргумент,
            в который будут передаваться параметры случившегося события в виде словаря(dict).

        :param event:           Имя события, для которого регистируется слушатель. Например:
                                    'DOM.attributeModified'.
        :param listener:        Колбэк-функция.
        :param args:            (optional) любое кол-во агрументов, которые будут переданы
                                    в функцию последними.
        :return:        None
        """
        e = event if type(event) is str else event.value
        if not iscoroutinefunction(listener):
            raise TypeError("Listener must be a async callable object!")
        if e not in self.listeners_for_event:
            self.listeners_for_event[ e ]: dict = {}
        self.listeners_for_event[ e ][listener] = args
        if not self.runtime_enabled:
            await self.Call("Runtime.enable")
            self.runtime_enabled = True

    def RemoveListenerForEvent(
            self, event: Union[str, DomainEvent], listener: Callable[[any], Awaitable[None]]) -> None:
        """
        Удаляет регистрацию слушателя для указанного события.
        :param event:           Имя метода, для которого была регистрация.
        :param listener:        Колбэк-функция, которую нужно удалить.
        :return:        None
        """
        e = event if type(event) is str else event.value
        if not iscoroutinefunction(listener):
            raise TypeError("Listener must be a async callable object!")
        if m := self.listeners_for_event.get( e ):
            if listener in m: m.pop(listener)


    def RemoveListenersForEvent(self, event: Union[str, DomainEvent]) -> None:
        """
        Удаляет регистрацию метода и слушателей вместе с ним для указанного события.
        :param event:          Имя метода, для которого была регистрация.
        :return:        None
        """
        e = event if type(event) is str else event.value
        if e in self.listeners_for_event:
            self.listeners_for_event.pop(e)

    def __del__(self) -> None:
        if self.verbose: log(f"[ DELETED ] {self.page_id}")



from abc import ABC, abstractmethod
from typing import Optional, Union
from ..Data import DomainEvent

class Log(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Log
    #   LogEntry -> https://chromedevtools.github.io/devtools-protocol/tot/Log#type-LogEntry
    """
    __slots__ = ()

    def __init__(self):
        self.log_domain_enabled = False

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def LogEnable(self) -> None:
        """
        Включает 'Log' домен, отправляет записи лога, собранные на данный момент, посредством
            события 'entryAdded'.
        https://chromedevtools.github.io/devtools-protocol/tot/Log#method-enable
        :return:
        """
        await self.Call("Log.enable")
        self.log_domain_enabled = True

    async def LogDisable(self) -> None:
        """
        Выключает 'Log' домен, останавливая отправку сообщений.
        https://chromedevtools.github.io/devtools-protocol/tot/Log#method-disable
        :return:
        """
        await self.Call("Log.disable")
        self.log_domain_enabled = False

    async def ClearLog(self) -> None:
        """
        Очищает список ранее опубликованных сообщений лога.
        https://chromedevtools.github.io/devtools-protocol/tot/Log#method-clear
        :return:
        """
        await self.Call("Log.clear")

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")

class LogEvent(DomainEvent):
    entryAdded = "Log.entryAdded"



import asyncio
from urllib.parse import quote
from abc import ABC, abstractmethod
from typing import Optional, Union, Callable, List, Literal
from ..Data import DomainEvent
from dataclasses import dataclass, field


class Page(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Page
    """
    __slots__ = ()

    def __init__(self):
        self.page_domain_enabled = False

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    @property
    def browser_name(self) -> str: return ""

    async def PageEnable(self) -> None:
        """
        Включает уведомления домена 'Page'.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-enable
        :return:
        """
        if not self.page_domain_enabled:
            await self.AddListenerForEvent("Page.frameStartedLoading", self._StateLoadWatcher, "started")
            await self.AddListenerForEvent("Page.frameNavigated", self._StateLoadWatcher, "navigated")
            await self.AddListenerForEvent("Page.frameStoppedLoading", self._StateLoadWatcher, "stopped")
            await self.Call("Page.enable")
            self.page_domain_enabled = True

    async def PageDisable(self) -> None:
        """
        Выключает уведомления домена 'Page'.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-disable
        :return:
        """
        if self.page_domain_enabled:
            self.RemoveListenerForEvent("Page.frameStartedLoading", self._StateLoadWatcher)
            self.RemoveListenerForEvent("Page.frameNavigated", self._StateLoadWatcher)
            self.RemoveListenerForEvent("Page.frameStoppedLoading", self._StateLoadWatcher)
            await self.Call("Page.disable")
            self.page_domain_enabled = False

    async def _StateLoadWatcher(self, params: dict, state: str) -> None:
        """
        Устанавливает состояние загрузки фрейма страницы, если включены уведомления
            домена Page.
        """
        frame_id = params["frameId"] if state != "navigated" else params["frame"]["id"]
        if frame_id == self.page_id:
            self.loading_state = state

    async def CreateIsolatedWorld(
            self, frameId: str, worldName: Optional[str] = None, grantUniversalAccess: Optional[bool] = None) -> int:
        """
        Создаёт изолированный мир для указанного фрейма.
        https://chromedevtools.github.io/devtools-protocol/1-3/Page/#method-createIsolatedWorld
        :param frameId:                 Идентификатор фрейма, в котором должен быть создан изолированный мир.
        :param worldName:               Необязательное имя, о котором сообщается в контексте выполнения.
        :param grantUniversalAccess:    Должен ли быть предоставлен универсальный доступ к изолированному миру.
                                            Это мощная опция, используйте ее с осторожностью.
        :return:            Runtime.ExecutionContextId
        """
        args = dict(frameId=frameId)
        if worldName is not None: args.update(worldName=worldName)
        if grantUniversalAccess is not None: args.update(grantUniversalAccess=grantUniversalAccess)
        result: dict = await self.Call("Page.createIsolatedWorld", args)
        return result.get("executionContextId")

    async def GetFrameTree(self) -> 'FrameTree':
        """
        Возвращает структуру присутствующих на странице фреймов. !!! Справедливо только в рамках одного домена.
        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-getFrameTree
        """
        result = await self.Call("Page.getFrameTree")
        # print("Page.getFrameTree", result)
        return FrameTree(**result.get("frameTree"))

    # async def GetFrameFor

    async def HandleJavaScriptDialog(self, accept: bool, promptText: str = "") -> None:
        """
        Подтверждает или закрывает диалоговое окно, инициированное JavaScript (alert, confirm,
            prompt или для onbeforeunload).
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-handleJavaScriptDialog
        :param accept:          'True' — принять, 'False' — отклонить диалог.
        :param promptText:      (optional) Текст для ввода в диалоговом окне, прежде чем принять.
                                    Используется, только если это диалоговое окно с подсказкой.
        :return:
        """
        args = {"accept": accept}
        if promptText: args.update({"promptText": promptText})
        await self.Call("Page.handleJavaScriptDialog", args)

    async def Navigate(
            self,
            url:  Union[str, bytes] = "about:blank",
            wait_for_load:     bool = True,
            wait_for_complete: bool = False
    ) -> None:
        """
        Переходит на адрес указанного 'url'.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-navigate
        :param url:                 Адрес, по которому происходит навигация.
        :param wait_for_load:       (optional) Если 'True' - ожидает 'complete' у 'document.readyState'
                                        страницы, на которую осуществляется переход, а так же состояния остановки
                                        газрузки ресурсов, если активны уведомления домена Page.
        :param wait_for_complete:   (optional) Если 'True' - ожидает 'complete' у 'document.readyState'
                                        страницы, на которую осуществляется переход.
        :return:
        """
        b_name_len = len(self.browser_name)
        is_url = (
            "http" == url[:4] or
            "chrome" == url[:6] or
            self.browser_name == url[:b_name_len] or
            url == "about:blank"
        )
        if self.page_domain_enabled: self.loading_state = "do_navigate"
        _url_ = ("data:text/html," + quote(url)
             # передать разметку как data-url, если начало этой строки
             # не содержит признаков url-адреса или передать "как есть",
             if type(url) is str and not is_url else url
                 # раз это строка содержащая url, или переход на пустую страницу
                 if type(url) is str and is_url else
                     # иначе декодировать и установить её как Base64
                     "data:text/html;Base64," + url.decode()
             )

        await self.Call("Page.navigate", {"url": _url_})
        if wait_for_load:
            await self.WaitForLoad(ignore_loading_state=wait_for_complete)

    async def WaitForLoad(
        self,
        desired_state:         str = "complete",
        interval:            float = .1,
        ignore_loading_state: bool = False
    ) -> None:
        """
        Дожидается указанного состояния загрузки документа.
            Если включены уведомления домена Page — дожидается, пока основной фрейм
            страницы не перестанет загружаться.
        :param desired_state:           (optional) Желаемое состояние загрузки. По умолчанию == полное.
        :param interval:                (optional) Таймаут ожидания.
        :param ignore_loading_state:    (optional) Ожидать только состояния загрузки страницы?
        :return:        None
        """

        if self.page_domain_enabled and not ignore_loading_state:
            while self.loading_state != "stopped":
                await asyncio.sleep(.1)
        else: await asyncio.sleep(1)

        while (await self.Eval("document.readyState"))["value"] != desired_state:
            await asyncio.sleep(interval)

    async def WaitNavigate(
            self,
            url: Union[str, bytes] = "about:blank",
            wait_for_load: bool    = True,
            timeout: float         = 10.0
    ) -> bool:
        """
        Дожидается совершения перехода на адрес за указанный интервал времени возвращая True, или False,
            если дождаться не удалось.
        :param url:                 Адрес, по которому происходит навигация.
        :param wait_for_load:       (optional) Если 'True' - ожидает 'complete' у 'document.readyState'
                                        страницы, на которую осуществляется переход.
        :param timeout:             (optional) Кол-во секунд ожидания.
        :return:
        """
        try: await asyncio.wait_for(self.Navigate(url, wait_for_load), timeout)
        except asyncio.exceptions.TimeoutError: return False
        else: return True

    async def AddScriptOnLoad(self, src: str) -> str:
        """
        Запускает полученный 'src' скрипта в каждом фрейме и перед загрузкой скриптов самого фрейма.
            Так же, такой скрипт будет запускаться автоматически в течении всей жизни инстанса,
            включая перезагрузку страницы.
        Требует включения( PageEnable() ) уведомлений домена "Page".
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-addScriptToEvaluateOnNewDocument
        :param src:             Текст сценария.
        :return:                identifier -> Уникальный идентификатор скрипта.
        """
        if not self.page_domain_enabled: await self.PageEnable()
        return (await self.Call("Page.addScriptToEvaluateOnNewDocument", {"source": src}))["identifier"]

    async def RemoveScriptOnLoad(self, identifier: str) -> None:
        """
        Удаляет данный скрипт из списка запускаемых при загрузке фрейма.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-removeScriptToEvaluateOnNewDocument
        :param identifier:      Идентификатор сценария.
        :return:
        """
        await self.Call("Page.removeScriptToEvaluateOnNewDocument", {"identifier": identifier})

    async def SetDocumentContent(self, html: str, frameId: str = None) -> None:
        """
        Устанавливает полученную разметку как HTML-код документа.
            frameId — можно найти среди параметров события 'Runtime.executionContextCreated',
            а так же у корневого элемента документа root.children[1].frameId
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setDocumentContent
        :param html:            HTML-разметка.
        :param frameId:         Идентификатор фрейма, которому предназначается html.
        :return:
        """
        if frameId is None: frameId = self.page_id
        await self.Call("Page.setDocumentContent", {"frameId": frameId, "html": html})

    async def SetAdBlockingEnabled(self, enabled: bool) -> None:
        """
        (EXPERIMENTAL)
        Включите экспериментальный рекламный фильтр Chrome на всех сайтах.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setAdBlockingEnabled
        :param enabled:         Включить?
        :return:
        """
        if not self.page_domain_enabled: await self.PageEnable()
        await self.Call("Page.setAdBlockingEnabled", {"enabled": enabled})

    async def SetFontFamilies(self, fontFamilies: dict) -> None:
        """
        (EXPERIMENTAL)
        Установить общие семейства шрифтов.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setFontFamilies
        :param fontFamilies:    Семейства шрифтов:
                                    {
                                        "standard":   str(), -> (optional)
                                        "fixed":      str(), -> (optional)
                                        "serif":      str(), -> (optional)
                                        "sansSerif":  str(), -> (optional)
                                        "cursive":    str(), -> (optional)
                                        "fantasy":    str(), -> (optional)
                                        "pictograph": str(), -> (optional)
                                    }
        :return:
        """
        await self.Call("Page.setFontFamilies", {"fontFamilies": fontFamilies})

    async def SetFontSizes(self, fontSizes: dict) -> None:
        """
        (EXPERIMENTAL)
        Установите размеры шрифта по умолчанию.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setFontSizes
        :param fontSizes:       Определяет размеры шрифта для установки. Если размер шрифта не
                                    указан, он не будет изменен:
                                    {
                                        "standard": int(), -> (optional)
                                        "fixed":    int(), -> (optional)
                                    }
        :return:
        """
        await self.Call("Page.setFontSizes", {"fontSizes": fontSizes})

    async def CaptureScreenshot(
        self,
        format_: str = "",
        quality: int = -1,
        clip: Optional[dict] = None,
        fromSurface: bool = True
    ) -> str:
        """
        Сделать скриншот. Возвращает кодированное base64 представление скриншота.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-captureScreenshot
        :param format_:         string => Image compression format (defaults to png) (jpeg or png).
        :param quality:         integer => Compression quality from range [0..100] (jpeg only).
        :param clip:            {
                                    "x": "number => X offset in device independent pixels (dip).",
                                    "y": "number => Y offset in device independent pixels (dip).",
                                    "width": "number => Rectangle width in device independent pixels (dip).",
                                    "height": "number => Rectangle height in device independent pixels (dip).",
                                    "scale": "number => Page scale factor."
                                }
        :param fromSurface:     boolean => Capture the screenshot from the surface, rather than the view.
                                    Defaults to true.
        :return:                string => Base64-encoded image data.
        """
        args = {"fromSurface": fromSurface}
        if format_: args.update({"format": format_})
        if quality > -1 and format_ == "jpeg": args.update({"quality": quality})
        if clip: args.update({"clip": clip})
        return (await self.Call("Page.captureScreenshot", args))["data"]

    async def PrintToPDF(
        self,
        landscape:               Optional[bool] = None,
        displayHeaderFooter:     Optional[bool] = None,
        printBackground:         Optional[bool] = None,
        scale:                  Optional[float] = None,
        paperWidth:             Optional[float] = None,
        paperHeight:            Optional[float] = None,
        marginTop:              Optional[float] = None,
        marginBottom:           Optional[float] = None,
        marginLeft:             Optional[float] = None,
        marginRight:            Optional[float] = None,
        pageRanges:               Optional[str] = None,
        ignoreInvalidPageRanges: Optional[bool] = None,
        headerTemplate:           Optional[str] = None,
        footerTemplate:           Optional[str] = None,
        preferCSSPageSize:       Optional[bool] = None,
        transferMode:             Optional[str] = None
    ) -> dict:
        """
        Печатает страницу как PDF.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-printToPDF
        :param landscape:               (optional) Ориентация бумаги. По умолчанию false.
        :param displayHeaderFooter:     (optional) Отобразить header и footer. По умолчанию false.
        :param printBackground:         (optional) Печать фоновой графики. По умолчанию false.
        :param scale:                   (optional) Масштаб рендеринга веб-страницы. По умолчанию 1.
        :param paperWidth:              (optional) Ширина бумаги в дюймах. По умолчанию 8,5 дюймов.
        :param paperHeight:             (optional) Высота бумаги в дюймах. По умолчанию 11 дюймов.
        :param marginTop:               (optional) Верхний отступ в дюймах. По умолчанию 1 см (~ 0,4 дюйма).
        :param marginBottom:            (optional) Нижний отступ в дюймах. По умолчанию 1 см (~ 0,4 дюйма).
        :param marginLeft:              (optional) Левый отступ в дюймах. По умолчанию 1 см (~ 0,4 дюйма).
        :param marginRight:             (optional) Правый отступ в дюймах. По умолчанию 1 см (~ 0,4 дюйма).
        :param pageRanges:              (optional) Диапазон бумаги для печати, например, '1-5, 8, 11-13'. По
                                            умолчанию используется пустая строка, что означает печать всех страниц.
        :param ignoreInvalidPageRanges: (optional) Следует ли игнорировать недействительные, но успешно
                                            проанализированные диапазоны страниц, например '3-2'. По умолчанию false.
        :param headerTemplate:          (optional) HTML-шаблон для печатного заголовка. Должна быть допустимая
                                            разметка HTML со следующими классами, используемыми для вставки
                                            значений печати в них:
                                                * date:       formatted print date
                                                * title:      document title
                                                * url:        document location
                                                * pageNumber: current page number
                                                * totalPages: total pages in the document
                                                    Например, <span class=title></span> будет генерировать span,
                                                        содержащий заголовок.
        :param footerTemplate:          (optional) HTML-шаблон для печати 'footer'. Следует использовать тот же
                                            формат, что и headerTemplate.
        :param preferCSSPageSize:       (optional) Предпочитать или нет размер страницы, как определено CSS. По
                                            умолчанию установлено значение false, и в этом случае содержимое будет
                                            масштабироваться по размеру бумаги.
        :param transferMode:            (optional, EXPERIMENTAL) вернуть как поток. Допустимые значения:
                                            ReturnAsBase64, ReturnAsStream
        :return:                        {
                                            "data": str(Данные PDF, кодированные в Base64.),
                                                ->  Будет пустым если в 'transferMode' выбрано 'ReturnAsStream'.
                                            "stream": str(Это либо получается из другого метода, либо указывается как blob <uuid> это UUID Blob.)
                                                -> StreamHandle
                                        }
        """
        args = {}
        if landscape is not None:               args.update({"landscape": landscape})
        if displayHeaderFooter is not None:     args.update({"displayHeaderFooter": displayHeaderFooter})
        if printBackground is not None:         args.update({"printBackground": printBackground})
        if scale is not None:                   args.update({"scale": scale})
        if paperWidth is not None:              args.update({"paperWidth": paperWidth})
        if paperHeight is not None:             args.update({"paperHeight": paperHeight})
        if marginTop is not None:               args.update({"marginTop": marginTop})
        if marginBottom is not None:            args.update({"marginBottom": marginBottom})
        if marginLeft is not None:              args.update({"marginLeft": marginLeft})
        if marginRight is not None:             args.update({"marginRight": marginRight})
        if pageRanges is not None:              args.update({"pageRanges": pageRanges})
        if ignoreInvalidPageRanges is not None: args.update({"ignoreInvalidPageRanges": ignoreInvalidPageRanges})
        if headerTemplate is not None:          args.update({"headerTemplate": headerTemplate})
        if footerTemplate is not None:          args.update({"footerTemplate": footerTemplate})
        if preferCSSPageSize is not None:       args.update({"preferCSSPageSize": preferCSSPageSize})
        if transferMode is not None:            args.update({"transferMode": transferMode})

        return await self.Call("Page.printToPDF", args)

    async def Reload(
        self,
        ignoreCache: bool = False,
        scriptToEvaluateOnLoad: str = "",
        wait_for_load: bool = True
    ) -> None:
        """
        Перезагружает страницу инстанса, при необходимости игнорируя кеш.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-reload
        :param ignoreCache:             (optional) Игнорировать кеш?
        :param scriptToEvaluateOnLoad:  (optional) Если установлено, сценарий будет вставлен во все
                                            фреймы проверяемой страницы после перезагрузки. Аргумент
                                            будет игнорироваться при перезагрузке источника dataURL.
        :param wait_for_load:           (optional) По умолчанию — дожидается полного завершения
                                            загрузки страницы(document.readyState === "complete").
                                            Установите False, если это поведение не требуется.
        :return:
        """
        if self.page_domain_enabled: self.loading_state = "do_reload"
        args = {}
        if ignoreCache:            args.update({"ignoreCache": ignoreCache})
        if scriptToEvaluateOnLoad: args.update({"scriptToEvaluateOnLoad": scriptToEvaluateOnLoad})
        await self.Call("Page.reload", args)
        if wait_for_load:
            await self.WaitForLoad()

    async def GetNavigationHistory(self) -> dict:
        """
        Возвращает историю навигации для текущей страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-getNavigationHistory
        :return:            {
                                "currentIndex": int(), -> Индекс текущей записи истории навигации.
                                "entries": list({
                                    "id": int(),  -> Уникальный идентификатор записи истории навигации.
                                    "url": str(), -> URL записи истории навигации.
                                    "userTypedURL": str(), -> URL, который пользователь ввел в строке URL.
                                    "title": str(), -> Название записи истории навигации.
                                    "transitionType": str(
                                        link, typed, address_bar, auto_bookmark, auto_subframe,
                                        manual_subframe, generated, auto_toplevel, form_submit,
                                        reload, keyword, keyword_generated, other
                                    ) -> Тип перехода.
                                }, ... ) -> Список записей истории навигации.
                            }
        """
        return await self.Call("Page.getNavigationHistory")

    async def NavigateToHistoryEntry(self, entryId: int) -> None:
        """
        Навигация текущей страницы к выбранной записи в истории навгации.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-navigateToHistoryEntry
        :param entryId:             Уникальный идентификатор записи для перехода.
        :return:
        """
        if self.page_domain_enabled: self.loading_state = "do_navigate"
        await self.Call("Page.navigateToHistoryEntry", {"entryId": entryId})

    async def ResetNavigationHistory(self) -> None:
        """
        Сбрасывает историю навигации для текущей страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-resetNavigationHistory
        :return:
        """
        await self.Call("Page.resetNavigationHistory")

    async def SetInterceptFileChooserDialog(self, enabled: bool) -> None:
        """
        Перехватывает запросы выбора файлов и передает управление клиентам протокола. Когда включен перехват
            файлов, диалоговое окно выбора файлов не отображается. Вместо этого генерируется событие
            протокола Page.fileChooserOpened.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setInterceptFileChooserDialog
        :param enabled:             Включить перехват?
        :return:
        """
        await self.Call("Page.setInterceptFileChooserDialog", {"enabled": enabled})

    @abstractmethod
    async def AddListenerForEvent(
            self, event: Union[str, DomainEvent], listener: Callable, *args: any) -> None:
        raise NotImplementedError("async method AddListenerForEvent() — is not implemented")

    @abstractmethod
    def RemoveListenerForEvent(self, event: str, listener: Callable) -> None:
        raise NotImplementedError("method RemoveListenerForEvent() — is not implemented")

    @abstractmethod
    async def Eval(
            self, expression: str,
            objectGroup:            str = "console",
            includeCommandLineAPI: bool = True,
            silent:                bool = False,
            returnByValue:         bool = False,
            userGesture:           bool = True,
            awaitPromise:          bool = False
    ) -> dict:
        raise NotImplementedError("async method Eval() — is not implemented")

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")


class PageEvent(DomainEvent):
    domContentEventFired = "Page.domContentEventFired"
    fileChooserOpened = "Page.fileChooserOpened"
    frameAttached = "Page.frameAttached"
    frameDetached = "Page.frameDetached"
    frameNavigated = "Page.frameNavigated"
    interstitialHidden = "Page.interstitialHidden"
    interstitialShown = "Page.interstitialShown"
    javascriptDialogClosed = "Page.javascriptDialogClosed"
    javascriptDialogOpening = "Page.javascriptDialogOpening"
    lifecycleEvent = "Page.lifecycleEvent"
    loadEventFired = "Page.loadEventFired"
    windowOpen = "Page.windowOpen"
    frameClearedScheduledNavigation = "Page.frameClearedScheduledNavigation"    # DEPRECATED
    frameScheduledNavigation = "Page.frameScheduledNavigation"                  # DEPRECATED
    backForwardCacheNotUsed = "Page.backForwardCacheNotUsed"                    # ! EXPERIMENTAL
    compilationCacheProduced = "Page.compilationCacheProduced"                  # ! EXPERIMENTAL
    documentOpened = "Page.documentOpened"                                      # ! EXPERIMENTAL
    frameRequestedNavigation = "Page.frameRequestedNavigation"                  # ! EXPERIMENTAL
    frameResized = "Page.frameResized"                                          # ! EXPERIMENTAL
    frameStartedLoading = "Page.frameStartedLoading"                            # ! EXPERIMENTAL
    frameStoppedLoading = "Page.frameStoppedLoading"                            # ! EXPERIMENTAL
    navigatedWithinDocument = "Page.navigatedWithinDocument"                    # ! EXPERIMENTAL
    screencastFrame = "Page.screencastFrame"                                    # ! EXPERIMENTAL
    screencastVisibilityChanged = "Page.screencastVisibilityChanged"            # ! EXPERIMENTAL
    downloadProgress = "Page.downloadProgress"                                  # * EXPERIMENTAL DEPRECATED
    downloadWillBegin = "Page.downloadWillBegin"                                # * EXPERIMENTAL DEPRECATED


@dataclass
class AdFrameStatus:
    """ Указывает, был ли frame идентифицирован как реклама и почему. """
    adFrameType: Literal["none", "child", "root"]
    explanations: Optional[List[Literal["ParentIsAd", "CreatedByAdScript", "MatchedBlockingRule"]]] = None


@dataclass
class Frame:
    """ Информация о фрейме на странице """
    id: str
    loaderId: str                               # ? Network.LoaderId
    url: str
    domainAndRegistry: str
    securityOrigin: str
    mimeType: str
    secureContextType: Literal["Secure", "SecureLocalhost", "InsecureScheme", "InsecureAncestor"]
    crossOriginIsolatedContextType: Literal["Isolated", "NotIsolated", "NotIsolatedFeatureDisabled"]
    gatedAPIFeatures: List[Literal[
        "SharedArrayBuffers", "SharedArrayBuffersTransferAllowed", "PerformanceMeasureMemory", "PerformanceProfile"
    ]]
    adFrameStatus: Optional['AdFrameStatus']
    parentId: Optional[str] = None
    name: Optional[str] = None
    urlFragment: Optional[str] = None
    unreachableUrl: Optional[str] = None
    _adFrameStatus: Optional['AdFrameStatus'] = field(init=False, repr=False, default=None)

    @property
    def adFrameStatus(self) -> Union['AdFrameStatus', None]:
        return self._adFrameStatus

    @adFrameStatus.setter
    def adFrameStatus(self, data: dict) -> None:
        self._adFrameStatus = AdFrameStatus(**data) if not isinstance(data, property) else None


@dataclass
class FrameTree:
    """ Информация об иерархии фреймов """
    frame: 'Frame'
    childFrames: Optional[List['FrameTree']]
    _childFrames: Optional[List['FrameTree']] = field(init=False, repr=False, default=None)

    @property
    def frame(self) -> 'Frame':
        return self._frame

    @frame.setter
    def frame(self, data: dict) -> None:
        self._frame = Frame(**data)

    @property
    def childFrames(self) -> Union[List['FrameTree'], None]:
        return self._childFrames

    @childFrames.setter
    def childFrames(self, data: List[dict]) -> None:
        if not isinstance(data, property):
            self._childFrames = [FrameTree(**frame_data) for frame_data in data]
        else:
            self._childFrames = None

    def GetFrameByUrl(self, value: str) -> Union['Frame', None]:
        def search(tree: 'FrameTree', value: str) -> Union['Frame', None]:
            print("tree.frame.url", tree.frame.url)
            if value in tree.frame.url:
                return tree.frame
            else:
                if tree.childFrames:
                    for child in tree.childFrames:
                        if result := search(child, value):
                            return result
            return None
        return search(self, value)



from setuptools import setup, find_packages
import re
from pathlib import Path


NAME = "aio_dt_protocol"
CURRENT_DIR = Path(__file__).parent
DIST_DIR = CURRENT_DIR / "build"
INIT_FILE_PATH = CURRENT_DIR / NAME / "__init__.py"

with open(INIT_FILE_PATH, encoding="utf-8") as f:
    INIT_CONTENT = f.read()

VERSION: str = re.findall(r"__version__[^\"']+\"([^\"']+)", INIT_CONTENT)[0]
AUTHOR: str = re.findall(r"__author__[^\"']+\"([^\"']+)", INIT_CONTENT)[0]
EMAIL: str = re.findall(r"__email__[^\"']+\"([^\"']+)", INIT_CONTENT)[0]

PACKAGES = find_packages()

with open("README.md", encoding="utf-8") as f:
    DESCRIPTION = f.read()


setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    description="Asynchronous wrapper over Chromium browser debugger protocol.",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/PieceOfGood/aio_dt_protocol",
    license="BSD 3-Clause",
    packages=PACKAGES,
    install_requires=["websockets>=10.0.0"],

    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX"
    ]
)



try:
    import ujson as json
except ModuleNotFoundError:
    import json

from .Page import Page
from .Actions import Actions
from .DOMElement import Node
from .Data import ViewportRect, WindowRect, GeoInfo

import asyncio
import base64, re
from typing import Optional, Union

from .exceptions import EvaluateError, JavaScriptError, NullProperty

from .domains.BackgroundService import BackgroundService as BackgroundServiceDomain
from .domains.Browser import Browser as BrowserDomain
from .domains.DOM import DOM as DOMDomain
from .domains.Emulation import Emulation as EmulationDomain
from .domains.Log import Log as LogDomain
from .domains.Network import Network as NetworkDomain
from .domains.Page import Page as PageDomain
from .domains.Runtime import Runtime as RuntimeDomain
from .domains.Target import Target as TargetDomain
from .domains.Console import Console as ConsoleDomain
from .domains.Overlay import Overlay as OverlayDomain
from .domains.CSS import CSS as CSSDomain
from .domains.DeviceOrientation import DeviceOrientation as DeviceOrientationDomain
from .domains.Fetch import Fetch as FetchDomain
from .domains.SystemInfo import SystemInfo as SystemInfoDomain

class PageEx(
    Page, BrowserDomain, DOMDomain, EmulationDomain, LogDomain, NetworkDomain,
    PageDomain, RuntimeDomain, TargetDomain, ConsoleDomain, OverlayDomain,
    CSSDomain, DeviceOrientationDomain, FetchDomain, SystemInfoDomain, BackgroundServiceDomain
):
    """
    Расширение для 'Page'. Включает сборку наиболее востребованных методов для работы
        с API 'ChromeDevTools Protocol', а так же дополняет некоторыми полезными методами.
    """
    __slots__ = (
        "ws_url", "page_id", "frontend_url", "callback", "is_headless_mode", "verbose", "browser_name", "id",
        "responses", "connected", "ws_session", "receiver", "on_detach_listener", "listeners", "listeners_for_method",
        "runtime_enabled", "on_close_event", "storage", "action", "_root", "style_sheets", "loading_state",
        "observing_started", "recording_started", "dom_domain_enabled", "targets_discovered", "log_domain_enabled",
        "network_domain_enabled", "console_domain_enabled", "page_domain_enabled", "fetch_domain_enabled",
        "css_domain_enabled", "overlay_domain_enabled"
    )

    def __init__(self, *args):
        Page.__init__(self, *args)

        BrowserDomain.__init__(self)
        DOMDomain.__init__(self)
        EmulationDomain.__init__(self)
        LogDomain.__init__(self)
        NetworkDomain.__init__(self)
        PageDomain.__init__(self)
        RuntimeDomain.__init__(self)
        TargetDomain.__init__(self)
        ConsoleDomain.__init__(self)
        OverlayDomain.__init__(self)
        CSSDomain.__init__(self)
        DeviceOrientationDomain.__init__(self)
        FetchDomain.__init__(self)
        SystemInfoDomain.__init__(self)
        BackgroundServiceDomain.__init__(self)

        self.storage = {}
        self.action = Actions(self)             # Совершает действия на странице. Клики; движения мыши; события клавиш
        self._root: Union[Node, None] = None
        self.style_sheets = []                  # Если домен CSS активирован, сюда попадут все 'styleSheetId' страницы

        self.loading_state = ""                 # Состояние загрузки страницы(основного фрейма). Отслеживается, если
                                                #   включены уведомления домена Page. Может иметь значения:
                                                # |  started; navigated; stopped; do_navigate; do_reload |

    def __eq__(self, other: "PageEx") -> bool:
        return self.page_id == other.page_id

    def __hash__(self) -> int:
        return hash(self.page_id)

    # region [ |>*<|=== Domains ===|>*<| ] Other [ |>*<|=== Domains ===|>*<| ]
    #

    async def PyExecAddOnload(self) -> None:
        """ Включает автоматически добавляющийся JavaScript, вызывающий слушателей
        клиента, добавленных на страницу с помощью await <Page>.AddListener(...) и
        await <Page>.AddListeners(...).

        Например, `test_func()` объявленная и добавленная следующим образом:

        async def test_func(number: int, text: str, bind_arg: dict) -> None:
            print("[- test_func -] Called with args:\n\tnumber: "
                  f"{number}\n\ttext: {text}\n\tbind_arg: {bind_arg}")

        await page.AddListener(
            test_func,                          # ! слушатель
            {"name": "test", "value": True}     # ! bind_arg
        )

        Может быть вызвана со страницы браузера, так:
        py_exec("test_func", 1, "testtt");
        """
        py_exec_js = """function py_exec(funcName, ...args) {
            console.info(JSON.stringify({ func_name: funcName, args: args })); }"""
        await self.AddScriptOnLoad(py_exec_js)

    async def GetViewportRect(self) -> ViewportRect:
        """
        Возвращает список с длиной и шириной вьюпорта браузера.
        """
        code = "(()=>{return JSON.stringify([window.innerWidth,window.innerHeight]);})();"
        data = json.loads(await self.InjectJS(code))
        return ViewportRect(int(data[0]), int(data[1]))

    async def GetWindowRect(self) -> WindowRect:
        """
        Возвращает список с длиной и шириной окна браузера.
        """
        code = "(()=>{return JSON.stringify([window.outerWidth,window.outerHeight]);})();"
        data = json.loads(await self.InjectJS(code))
        return WindowRect(int(data[0]), int(data[1]))

    async def GetUrl(self) -> str:
        return (await self.GetTargetInfo()).url

    async def GetTitle(self) -> str:
        return (await self.GetTargetInfo()).title

    async def MakeScreenshot(
            self,
                 format_: str = "",
                 quality: int = -1,
                   clip: Optional[dict] = None,
            fromSurface: bool = True
    ) -> bytes:
        """
        Сделать скриншот. Возвращает набор байт, представляющий скриншот.
        :param format_:         jpeg или png (по умолчанию png).
        :param quality:         Качество изображения в диапазоне [0..100] (только для jpeg).
        :param clip:            {
                                    "x": "number => X offset in device independent pixels (dip).",
                                    "y": "number => Y offset in device independent pixels (dip).",
                                    "width": "number => Rectangle width in device independent pixels (dip).",
                                    "height": "number => Rectangle height in device independent pixels (dip).",
                                    "scale": "number => Page scale factor."
                                }
        :param fromSurface:     boolean => Capture the screenshot from the surface, rather than the view.
                                    Defaults to true.
        :return:                bytes
        """
        shot = await self.CaptureScreenshot(format_, quality, clip, fromSurface)
        return base64.b64decode(shot.encode("utf-8"))

    async def SelectInputContentBy(self, css: str) -> None:
        await self.InjectJS(f"let _i_ = document.querySelector('{css}'); _i_.focus(); _i_.select();")

    async def ScrollIntoViewJS(self, selector: str) -> None:
        await self.InjectJS(
            "document.querySelector(\"" +
            selector +
            "\").scrollIntoView({'behavior':'smooth', 'block': 'center'});"
        )

    async def InjectJS(self, code: str) -> any:
        """ Выполняет JavaScript-выражение во фрейме верхнего уровня. """
        try:
            result = await self.Eval(code)
        except EvaluateError as error:
            error = str(error)
            if "of null" in error:
                if match := re.match(r"[\w\s:]+['|\"]([^'\"]+)", error):
                    prop = match.group(1)
                else:
                    prop = "unmatched error: " + error
                raise NullProperty(f"InjectJS() Exception with injected code:\n'{code}'\nNull property:\n{prop}")

            raise JavaScriptError(f"JavaScriptError: InjectJS() Exception with injected code:\n'{code}'\nDescription:\n{error}")

        return result.get('value')

    async def GetGeoInfo(self) -> GeoInfo:
        """
        Возвращает информацию о Вашем местоположении, вычисленному по IP.
        """
        async_fn_js = """\
        async function get_geo_info() {
            const resp = await fetch('https://time.gologin.com/');
            return await resp.text();
        } get_geo_info();
        """

        promise = """fetch('https://time.gologin.com/').then(res => res.text())"""

        result: dict = await self.EvalPromise(promise)
        result.update(
            geo=dict(
                latitude=float(result["ll"][0]),
                longitude=float(result["ll"][1]),
                accuracy=float(result["accuracy"])
            ),
            languages=result["languages"].split(","),
            state_province=result.get("stateProv"),
            proxy_type=(pt:=result.get("proxyType"))
        )
        del result["ll"]
        del result["accuracy"]
        del result["stateProv"]
        if pt is not None:
            del result["proxyType"]
        return GeoInfo(**result)


    async def CatchMetaForUrl(self, url: str, uniq_key: Optional[str] = None) -> None:
        """
        Получает заголовки запроса, ответа, а так же cookie для конкретного url и сохраняет их
            в виде словарей в поле собственного инстанса, доступного как:
            page_instance.storage[uniq_key]. Если 'uniq_key' не указан, в его качестве будет
            использован 'url'.

        Для работы будет активирован домен Fetch со значением "True" для "fetch_onResponse".
        :param url:             Адрес ресурса, или его уникальная часть.
        :param uniq_key:        Идентификатор, под которым будут сохраняться данные перехвата.
        :return:
        """
        if not self.fetch_domain_enabled:
            await self.FetchEnable()
        key = uniq_key if uniq_key else url
        await self.AddListenerForEvent("Fetch.requestPaused", catch_headers_for_url, self, url, self.storage, key)


    # endregion

async def catch_headers_for_url(data: dict, instance: 'PageEx', url: str, storage: dict, storage_key: str) -> None:
    """
    Получает заголовки запроса, ответа, а так же cookie для конкретного url и сохраняет их
    в виде словарей в переданном по ссылке 'storage'. Требует активации домена Fetch со значением
    "True" для "fetch_onResponse", после чего, инстансу страницы, эта корутина добавляется как
     слушатель транслирующий через 'data' параметры перехваченных запросов. Например:
            await page.FetchEnable( fetch_onResponse=True )
            await page.AddListenerForEvent(
                "Fetch.requestPaused",
                catch_headers_for_url,
                page,
                "persistentBadging",
                GLOBAL_STORAGE,
                "storage_key"
            )
    Структура сохранённых данных: (dict) {
        "request_url":      (str) - полный url перехваченного запроса,
        "request_headers":  (dict) - {"hdr_name": "hdr_value", "hdr_name": "hdr_value", ... },
        "response_headers": (dict) - {"hdr_name": "hdr_value", "hdr_name": "hdr_value", ... },
        "response_cookies": (dict) - {"cook_name": "cook_value", "cook_name": "cook_value", ... },
        "full_cookie_data": (list[ dict ]) - [
            {
                "name": "cookie name",
                "value": "cookie value",
                "domain": ".instagram.com",
                "path": "/",
                "expires": -1,
                "size": 149,
                "httpOnly": true,
                "secure": true,
                "session": true,
                "priority": "Medium"
            }, { ... }
        ]
    }
    :param data:            Содержимое перехваченного запроса. Эти данные будут переданы в слушатель из браузера.
                                Структура: {
                                    requestId: (str),
                                    request: {
                                        url: (str),
                                        method: (str),
                                        headers: (dict) — Все заголовки запроса включая куки,
                                        postData: (str) — payload POST-запросов,
                                        initialPriority: (str),
                                        referrerPolicy: (str),
                                        Возможно какие-то ещё ...
                                    },
                                    frameId: (str) — like "75ED3620550E8F5E16B8F6D1FF2EE4D1",
                                    resourceType: (str) — like "XHR",
                                    responseStatusCode: (int) — like 200,
                                    responseHeaders: (List[dict]) — заголовки ответа, не включая куки:
                                        [ {'name': 'content-type', 'value': '"application/json; charset=utf-8'}, ... ]
                                }
    :param instance:        Инстанс страницы, от лица которой происходит перехват.
    :param url:             Адрес ресурса, или его уникальная часть.
    :param storage:         Ссылка на словарь, в который будет сохранён результат.
    :param storage_key:     Идентификатор, под которым сохранится запись в 'storage'.
    :return:    None
    """
    asyncio.create_task(instance.ContinueRequest(data["requestId"]))
    if url in data["request"]["url"]:
        response_headers = {}; response_cookies = {}
        for item in data["responseHeaders"]: response_headers[item["name"]] = item["value"]
        full_cookie_data = await instance.GetCookies()
        for item in full_cookie_data:
            response_cookies[item["name"]] = item["value"]
        storage[storage_key] = {
            "request_url":      data["request"]["url"],
            "request_headers":  data["request"]["headers"],
            "response_headers": response_headers,
            "response_cookies": response_cookies,
            "full_cookie_data": full_cookie_data
        }
